/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : PathBuilder.h                                            */
/*                                                                       */
/*       Date : September 20, 1998                                       */
/* Updated at : February  15, 1999                                       */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <stdlib.h>
#include <iostream.h>

#include <WeightedRandomPickerSet.h>
#include <RandomGenerator.h>

#include "ListOfEdgeLists.h"
#include "EdgeGraph.h"
#include "PathBuilder.h"
#include "Flight.h"

/*************************************************************************/
/**
    DFS cycle finder.
*/
/*************************************************************************/

PathBuilder::lookAheadForCycles( EdgeGraph& eg, Edge** cstack, 
				 int& deep, int maxdeep, Flight* fflight,
				 float maxLength )
{
   Edge* peek = cstack[deep]; 
   int startCityIndex = fflight->getFromIndex();
   Flight *dflight;

   if(peek==NULL) return 0;

   Edge* tmp;
   int ret=0, i, j;
   WeightedRandomPickerSet<Edge*> leaving(10);

   if( eg.getLeavingEdges(peek->getEndNode(),&leaving) )
   { 
     int size = leaving.getSize();
     j = RandomGenerator::generateInt(0,size); 
     for(i=0;i<leaving.getSize()&&!ret;i++)
     {
	// It should start and end at the same base city.
	// It should not be longer than the allowed rotation time.
	if( leaving.getElementAtSafe((i+j)%size,tmp) && 
	    tmp->getDestinationCityIndex()==startCityIndex &&
	    (dflight=myProblem->getFlightWithIndex(tmp->getEndNode()))!=NULL && 
	    (dflight->getArrivalDate()-fflight->getDepartureDate())<=maxLength)
        { ret=1; deep++; cstack[deep]=tmp; }

     }// end of for

     if( !ret )
     {
       j =  RandomGenerator::generateInt(0,size); //rand() % size;
       for(i=0;i<leaving.getSize()&&!ret;i++)	
        if( leaving.getElementAtSafe((i+j)%size,tmp) )
        {
          if((dflight=myProblem->getFlightWithIndex(tmp->getEndNode()))!=NULL&&
             (dflight->getArrivalDate()-fflight->getDepartureDate())<=maxLength)
          {
	     deep++;
	     cstack[deep]=tmp;
             ret = lookAheadForCycles(eg,cstack,deep,maxdeep,fflight,maxLength);
             if(!ret) deep--; 

          }// end of if

        }// end of if

     }// end of if

   }// end of if

   return ret;

}// end of lookAheadForCycles()

/*************************************************************************/
/**
    1. For the edge set (candids), try to find out a cycle no more than
       3 edges. It will collect all these into 'retCycles'.
       APPLYBASECITY controls whether BaseCity constraint will be forced
       or not.
       APPLYBASECITY = 1 : Force
       APPLYBASECITY = 0 : Relax
*/
/*************************************************************************/

void 
PathBuilder::searchForwardWithEdgeSet( EdgeGraph& eg, 
				       WeightedRandomPickerSet<Edge*>& candids, 
				       ListOfEdgeLists& retCycles, 
				       ListOfEdgeLists& paths, 
				       float maxLength,	
				       int APPLYBASECITY)
{
   Edge* tmp;
   Edge** cstack = new Edge*[eg.getAirProblem()->getNumberOfFlights()];

   int i, j, notfound, size;
   int deep=0;

   size = candids.getSize();

   if( size ) 
   {
      j =  RandomGenerator::generateInt(0,size); 
      notfound=1;

      for(i=0;notfound&&i<size;i++)
	if( candids.getElementAtSafe((i+j)%size,tmp) && tmp!=NULL )
        {

	    // It might be the case that this is a short trip.  
	    if( tmp->getDestinationCityIndex() == tmp->getStartCityIndex() &&
		( tmp->getEndDate()-tmp->getStartDate() ) <= maxLength )
	    {
	       EdgeList e;
	       e.addEdge(tmp);

	       if(e.isValid(myProblem))
	       {	 
	          if( APPLYBASECITY )
	          { 
		     if( tmp->doesStartFromBaseCity() ) 
		          retCycles.addEdgeList(e);
	             else paths.addEdgeList(e);

	          } else retCycles.addEdgeList(e);

	          notfound=0;

	       }// end of if 

	    }// end of if

	    deep=0;
	    cstack[deep]=tmp;

	    if( lookAheadForCycles(eg,cstack, deep, 3, tmp->getFirstFlight(), maxLength ) )
	    {
	       EdgeList e;

               int ok=1,m=deep;
	       while( m>=0 && ok ) ok = e.addEdgeFront(cstack[m--]);

	       if( ok && e.isValid(myProblem) )
	       {

                  // If Base City constraint is not forced, then 
		  // any cycle can be returned. Otherwise, we need to check that
		  // the cycle starts from and ends at the same base city.
	          if( APPLYBASECITY )
	          { 
		    if( tmp->doesStartFromBaseCity() ) 
		         retCycles.addEdgeList(e);
	            else paths.addEdgeList(e);

	          } else retCycles.addEdgeList(e);

	         notfound=0;

	       } //else cstack.clear();

	    }// end of if

	}// end of if

   }// end of if

   delete[] cstack;

}// end of searchForwardWithEdgeSet()

/*************************************************************************/
/**
    1. find cycles starting from incoming edges.
    2. find cycles starting from leaving edges.
    3. if we can not produce anything so far, then look at paths:
       3.a. Paths starting from the leaving endges. If the current
	    flight starts from Basecity then, it can be added to the
	    schedule.
       3.b. Paths starting from incoming edges if they start from
	    base city.
*/
/*************************************************************************/

void 
PathBuilder::buildPaths( EdgeGraph& eg, ListOfEdgeLists& cycles, 
			 ListOfEdgeLists& paths, int FlightNumber, 
			 float maxLength)
{
   int i,j, size;
   Edge* tmp;
   WeightedRandomPickerSet<Edge*> candid1 = eg.getIncomingEdges(FlightNumber);
   WeightedRandomPickerSet<Edge*> candid2 = eg.getLeavingEdges(FlightNumber);

   // Investigate the paths forward fashion starting with incoming edges.
   if( candid1.getSize() ) 
       searchForwardWithEdgeSet(eg,candid1,cycles,paths,maxLength);

   // Investigate the paths forward fashion starting with leaving edges.
   if( candid2.getSize() ) 
       searchForwardWithEdgeSet(eg,candid2,cycles,paths,maxLength);

   // If there is no cycle or path anything like that, then we need to try to cover it
   // somehow by only one edge.
   if( !(cycles.getSize()+paths.getSize()) )
   {
      if( myProblem->isFlightFromBaseCity(FlightNumber) && candid2.getSize() )
      {
         // get the size of the set of leaving edges 
         size = candid2.getSize();

         // generate a random number between 0 and size 
         j=RandomGenerator::generateInt(0,size); 

         // add each element to the paths since we need to cover this node 
         // one way or another.
	 for(i=0;i<size;i++)
           if( candid2.getElementAtSafe((i+j)%size,tmp) && 
	       ( tmp->doesStartFromBaseCity() || tmp->doesEndAtBaseCity() ) ) 
	   { EdgeList e(tmp); paths.addEdgeList(e); }

      }// end of if

      if( candid1.getSize() && paths.getSize()<1 )
      {
         // get the size of the set of incoming edges 
         size = candid1.getSize();

         // generate a random number between 0 and size 
         j=RandomGenerator::generateInt(0,size); 

         // add each element to the paths since we need to cover this node 
         // one way or another.
        for(i=0;i<size;i++)
          if( candid1.getElementAtSafe((i+j)%size,tmp) && 
	      ( tmp->doesStartFromBaseCity() || tmp->doesEndAtBaseCity() ) ) 
          { EdgeList e(tmp); paths.addEdgeList(e); }

      }// end of if

   }// end of if
  
}// end of buildPaths()

/*************************************************************************/
/**
    DFS cycle finder.
*/
/*************************************************************************/

PathBuilder::lookBackForCycles( EdgeGraph& eg, Edge** cstack, 
			        int deep, int maxdeep, Flight *fflight,
				float maxLength )
{
   Edge* peek = cstack[deep];//*cstack.pop();
   Flight *dflight ;
   int endCityIndex = fflight->getToIndex();

   if(peek==NULL) return 0;

   Edge* tmp;
   WeightedRandomPickerSet<Edge*> incoming = eg.getIncomingEdges(peek->getStartNode());
   int ret=0,i,j;

   if( incoming.getSize() )
   {
     j=RandomGenerator::generateInt(0,incoming.getSize()); //rand()%incoming.getSize();
     for(i=0;i<incoming.getSize()&&!ret;i++)
     {
	// It should start and end at the same base city.
	// It should not be longer than the allowed rotation time.
	if( incoming.getElementAtSafe((i+j)%incoming.getSize(),tmp) && 
	    tmp->getStartCityIndex()==endCityIndex && 
	    (dflight=tmp->getFirstFlight())!=NULL &&
	    (fflight->getArrivalDate()-dflight->getDepartureDate())<=maxLength
	    )
        { ret=1; deep++; cstack[deep] = tmp; }

     }// end of for

     if( !ret )
     {
       j=RandomGenerator::generateInt(0,incoming.getSize()); 
       for(i=0;i<incoming.getSize()&&!ret;i++)	
         if( incoming.getElementAtSafe((i+j)%incoming.getSize(),tmp) &&
             (dflight=tmp->getFirstFlight())!=NULL &&
             (fflight->getArrivalDate()-dflight->getDepartureDate())<=maxLength )
         {
	     deep++; cstack[deep] = tmp;
	     ret = lookBackForCycles(eg,cstack,deep,maxdeep,fflight,maxLength);
	     if(!ret) deep--; 

	 }// end of if

     }// end of if

   }// end of if

   return ret;

}// end of lookBackForCycles()

/*************************************************************************/
/**
    1. For the edge set (candids), try to find out a cycle no more than
       3 edges. It will collect all these into 'retCycles'.
       APPLYBASECITY controls whether BaseCity constraint will be forced
       or not.
       APPLYBASECITY = 1 : Force
       APPLYBASECITY = 0 : Relax
*/
/*************************************************************************/

void 
PathBuilder::searchBackwardWithEdgeSet( EdgeGraph& eg, 
				       WeightedRandomPickerSet<Edge*>& candids, 
				       ListOfEdgeLists& retCycles, 
				       ListOfEdgeLists& paths, 
				       float maxLength,
				       int APPLYBASECITY)
{
   Edge* tmp;
   Edge** cstack = new Edge*[eg.getAirProblem()->getNumberOfFlights()];

   int i, j, size, notfound;
   int deep=0;

   size = candids.getSize();

   if( size )
   {
      notfound=1; 
      j = RandomGenerator::generateInt(0,size); 
      for(i=0;i<size&&notfound;i++)
	if( candids.getElementAtSafe((i+j)%size,tmp) && tmp!=NULL )
        {

	    if( tmp->getStartCityIndex()==tmp->getDestinationCityIndex() &&
		(tmp->getEndDate()-tmp->getStartDate()) <= maxLength )
            {
	       EdgeList e;
	       e.addEdge(tmp);

	       if( e.isValid(myProblem) )
	       {	
	          // Add to retCycles if it starts from and ends at the
	          // same base city.
	          if( APPLYBASECITY )
	          {
	             if( tmp->doesEndAtBaseCity() ) retCycles.addEdgeList(e);
	             else paths.addEdgeList(e);

	          } else retCycles.addEdgeList(e);

	          notfound=0;

	       }// end of if	
	    }

	    deep=0;
	    cstack[deep]=tmp;
	    if( lookBackForCycles(eg, cstack, deep, 3, tmp->getSecondFlight(), maxLength ))
	    {
	       EdgeList e;
	       int m=0;

	       while( m<=deep ) e.addEdge( cstack[m++] );

	       if( e.isValid(myProblem) )
	       {
	          // Add to retCycles if it starts from and ends at the
	          // same base city.
	          if( APPLYBASECITY )
	          {
	             if( tmp->doesEndAtBaseCity() ) retCycles.addEdgeList(e);
	             else paths.addEdgeList(e);

	          } else retCycles.addEdgeList(e);

	          notfound=0;

	       }// end of if

	    }// end of if

	}// end of if

   }// end of if

   delete[] cstack;

}// end of searchBackwardWithEdgeSet()

/*************************************************************************/
/**
    1. find cycles starting from incoming edges.
    2. find cycles starting from leaving edges.
    3. if we can not produce anything so far, then look at paths:
       3.a. Paths starting from the leaving endges. If the current
	    flight starts from Basecity then, it can be added to the
	    schedule.
       3.b. Paths starting from incoming edges if they start from
	    base city.
*/
/*************************************************************************/

void 
PathBuilder::buildPathsBackwards( EdgeGraph& eg, ListOfEdgeLists& cycles, 
			         ListOfEdgeLists& paths, int FlightNumber,
				 float maxLength )
{
   int i, j, size;
   Edge* tmp;
   WeightedRandomPickerSet<Edge*> candid1 = eg.getIncomingEdges(FlightNumber);
   WeightedRandomPickerSet<Edge*> candid2 = eg.getLeavingEdges(FlightNumber);

   // Investigate the paths backward fashion starting with incoming edges.
   if( candid1.getSize() ) 
       searchBackwardWithEdgeSet(eg,candid1,cycles,paths,maxLength);

   // Investigate the paths backward fashion starting with leaving edges.
   if( candid2.getSize() ) 
       searchBackwardWithEdgeSet(eg,candid2,cycles,paths,maxLength);

   // If there is no cycle or path anything like that, then we need to try to cover it
   // somehow by only one edge.
   if( !(cycles.getSize()+paths.getSize()) )
   {
      if( myProblem->isFlightToBaseCity(FlightNumber) && candid1.getSize() )
      {
	 // get the size of the set of incoming edges 
	 size = candid1.getSize();

	 // generate a random number between 0 and size 
	 j=RandomGenerator::generateInt(0,size); 

	 // add each element to the paths since we need to cover this node 
	 // one way or another.
	 for(i=0;i<candid1.getSize();i++)
	   if( candid1.getElementAtSafe((i+j)%size,tmp) && 
	       ( tmp->doesStartFromBaseCity() || tmp->doesEndAtBaseCity() ) )
	   { EdgeList e(tmp); if(e.isValid(myProblem)) paths.addEdgeList(e); }

      }// end of if

      if( candid2.getSize() && paths.getSize()<1 )
      {
	 // get the size of the set of leaving edges 
	 size = candid2.getSize();

	 // generate a random number between 0 and size 
	 j=RandomGenerator::generateInt(0,size); 

	 // add each element to the paths since we need to cover this node 
	 // one way or another.
         for(i=0;i<size;i++)
           if( candid2.getElementAtSafe((i+j)%size,tmp) && 
	       ( tmp->doesStartFromBaseCity() || tmp->doesEndAtBaseCity() ) )
           { EdgeList e(tmp); if(e.isValid(myProblem)) paths.addEdgeList(e); }

      }// end of if

   }// end of if
  
}// end of buildPathsBackward()

/*************************************************************************/
/*************************************************************************/

void PathBuilder::println( WeightedRandomPickerSet<Edge*>* _set )
{
   int i;
   Edge* tmp;

   cout << " set = { ";

   for(i=0;i<_set->getSize()-1;i++)
      if( _set->getElementAtSafe(i,tmp) ) 
      { Edge::print( tmp, cout ); cout << " , "; }// end of if

   if( _set->getElementAtSafe(i,tmp) ) Edge::print( tmp, cout ); 

   cout << " } " << endl;

}// end of println()

/*************************************************************************/
/**
*/
/*************************************************************************/

PathBuilder::searchForCycles( EdgeGraph& eg, 
			      Edge** cstack, 
			      ListOfEdgeLists& retCycles, 
		              int& first, int middle, int& last, 
		  	      float maxLength, int APPLYBASECITY )
{
   int numberOfFlights = eg.getAirProblem()->getNumberOfFlights();
   Edge* tmp;
   int   i, j, found = 0;
   Edge* firstE = cstack[first];
   Edge* lastE  = cstack[last];
   WeightedRandomPickerSet<Edge*> leaving(10);
   int isize = 0;

   if( firstE->getStartCityIndex()==lastE->getDestinationCityIndex() &&
       ( !APPLYBASECITY || firstE->doesStartFromBaseCity() ) )
   {
      EdgeList e(numberOfFlights);

      int ok=1,m=last;

      while( m>=first && ok ) { ok = e.addEdgeFront(cstack[m]); m--; }

      if( ok && e.isValid(myProblem) )
      {   retCycles.addEdgeList(e); found=1; } 
      
   }// end of if 

   if( !found )
   {

     WeightedRandomPickerSet<Edge*> incoming = eg.getIncomingEdges(firstE->getStartNode());
     isize = incoming.getSize();

     if( isize )
     {
       j=RandomGenerator::generateInt(0,isize); 
       for(i=0;i<isize&&!found;i++)
       {
	 // It should not be longer than the allowed rotation time.
	 if( incoming.getElementAtSafe((i+j)%isize,tmp) && 
	     (lastE->getEndDate()-tmp->getStartDate())<=maxLength && 
	     first>0 )
         { 
	     first--;
	     cstack[first] = tmp; 
	     if(searchForCycles(eg,cstack,retCycles,first,middle,last, maxLength,APPLYBASECITY)) found = 1;
             first++;

	 }// end of if

       }// end of for

     }// end of if

   }// end of if

   if( !found && eg.getLeavingEdges(lastE->getEndNode(),&leaving) )
   {

     isize = leaving.getSize();

     if( isize ) 
     {
       j=RandomGenerator::generateInt(0,isize); 
       for(i=0;i<isize&&!found;i++)
       {
	 // It should not be longer than the allowed rotation time.
	 if( leaving.getElementAtSafe((i+j)%isize,tmp) && 
	     (tmp->getEndDate()-firstE->getStartDate())<=maxLength && 
	     last<=2*middle && last>=middle )
         { 
	     last++;
	     cstack[last] = tmp; 
	     if(searchForCycles(eg,cstack,retCycles,first,middle,last, maxLength,APPLYBASECITY)) found = 1;
             last--;

	 }// end of if

       }// end of for

     }// end of if

   }// end of if

   return found;

}// end of searchForCycles()

/*************************************************************************/
/**
    1. For the edge set (candids), try to find out a cycle no more than
       3 edges. It will collect all these into 'retCycles'.
       APPLYBASECITY controls whether BaseCity constraint will be forced
       or not.
       APPLYBASECITY = 1 : Force
       APPLYBASECITY = 0 : Relax
*/
/*************************************************************************/

void 
PathBuilder::searchBothWithEdgeSet( EdgeGraph& eg, 
		                    WeightedRandomPickerSet<Edge*>& candids, 
			            ListOfEdgeLists& retCycles, 
			            float maxLength,	
				    int numOfSolutions, 
			            int APPLYBASECITY)
{
   int numberOfFlights = eg.getAirProblem()->getNumberOfFlights();
   Edge*  tmp;
   Edge** cstack = new Edge*[2*numberOfFlights+1];
   int    first, last, middle;

   int i, j, notfound, size;
   int deep=0;

   size = candids.getSize();

   if( size ) 
   {
      j =  RandomGenerator::generateInt(0,size); 
      notfound=1;

      for(i=0;retCycles.getNumberOfEdgeLists()<numOfSolutions && i<size ;i++)
	if( candids.getElementAtSafe((i+j)%size,tmp) && tmp!=NULL )
        {

	    notfound = 1;

	    // It might be the case that this is a short trip.  
	    if(   tmp->getDestinationCityIndex() == tmp->getStartCityIndex() &&
		( tmp->getEndDate()-tmp->getStartDate() ) <= maxLength &&
	        ( !APPLYBASECITY || tmp->doesStartFromBaseCity() ) ) 
	    {
	       EdgeList e(numberOfFlights);
	       e.addEdge(tmp);

	       if(e.isValid(myProblem)) 
	       { retCycles.addEdgeList(e); notfound = 0; }

	    }// end of if

	    // If we already find a short trip then there is no need to
 	    // go forward.
	    if( notfound && retCycles.getNumberOfEdgeLists() < numOfSolutions )
	    { 
	      cstack[numberOfFlights]=tmp;
	      first = last = numberOfFlights;

	      //if( notfound )
              //notfound = 
	      searchForCycles(eg,cstack,retCycles,first, numberOfFlights,last,
			      maxLength,APPLYBASECITY) ;

	    }// end of if 

	}// end of if

   }// end of if

   delete[] cstack;

}// end of searchBothWithEdgeSet()

/*************************************************************************/
/**
    eg : Current Edge Graph
    cycles : Return value
    FlightNumber : The flight needs to be covered
    maxLength : Maximum length of a cycle
    APPLYBASECITY : Base city constrained is applied.
*/
/*************************************************************************/
void 
PathBuilder::buildCycles( EdgeGraph& eg, ListOfEdgeLists& cycles, 
			  int FlightNumber, float maxLength, int APPLYBASECITY )
{
   int i, j, size;
   Edge* tmp;
   WeightedRandomPickerSet<Edge*> candid1 = eg.getIncomingEdges(FlightNumber);
   WeightedRandomPickerSet<Edge*> candid2 = eg.getLeavingEdges(FlightNumber);
   int toggle = RandomGenerator::generateInt(0,2); 

   if(toggle)
   {

      // Investigate the paths backward fashion starting with incoming edges.
      if( candid1.getSize() ) 
       searchBothWithEdgeSet(eg,candid1,cycles,maxLength,2,APPLYBASECITY);

      // Investigate the paths backward fashion starting with leaving edges.
      if( cycles.getNumberOfEdgeLists()==0 && candid2.getSize() ) 
       searchBothWithEdgeSet(eg,candid2,cycles,maxLength,2,APPLYBASECITY);

   } 
   else
    {

      // Investigate the paths backward fashion starting with incoming edges.
      if( candid2.getSize() ) 
       searchBothWithEdgeSet(eg,candid2,cycles,maxLength,2,APPLYBASECITY);

      // Investigate the paths backward fashion starting with leaving edges.
      if( cycles.getNumberOfEdgeLists()==0 && candid1.getSize() ) 
       searchBothWithEdgeSet(eg,candid1,cycles,maxLength,2,APPLYBASECITY);
    } 

}// end of buildCycles()

/*************************************************************************/
/**
    eg : Current Edge Graph
    cycles : Return value
    FlightNumber : The flight needs to be covered
    maxLength : Maximum length of a cycle
    APPLYBASECITY : Base city constrained is applied.
*/
/*************************************************************************/
void 
PathBuilder::buildCyclesN( EdgeGraph& eg, ListOfEdgeLists& cycles, 
			  int FlightNumber, float maxLength, 
			  int numOfSolutions, int APPLYBASECITY )
{
   int i, j, size;
   Edge* tmp;
   WeightedRandomPickerSet<Edge*> candid1 = eg.getIncomingEdges(FlightNumber);
   WeightedRandomPickerSet<Edge*> candid2 = eg.getLeavingEdges(FlightNumber);
   int toggle = RandomGenerator::generateInt(0,2); 

   if(toggle)
   {

      // Investigate the paths backward fashion starting with incoming edges.
      if( candid1.getSize() ) 
       searchBothWithEdgeSet(eg,candid1,cycles,maxLength,numOfSolutions,APPLYBASECITY);

      // Investigate the paths backward fashion starting with leaving edges.
      if( cycles.getNumberOfEdgeLists() < numOfSolutions && candid2.getSize() ) 
       searchBothWithEdgeSet(eg,candid2,cycles,maxLength,numOfSolutions,APPLYBASECITY);

   } 
   else
    {

      // Investigate the paths backward fashion starting with incoming edges.
      if( candid2.getSize() ) 
       searchBothWithEdgeSet(eg,candid2,cycles,maxLength,numOfSolutions,APPLYBASECITY);

      // Investigate the paths backward fashion starting with leaving edges.
      if( cycles.getNumberOfEdgeLists() < numOfSolutions && candid1.getSize() ) 
       searchBothWithEdgeSet(eg,candid1,cycles,maxLength,numOfSolutions,APPLYBASECITY);
    } 

}// end of buildCyclesN()

